<?php

namespace TotalSurvey\Extensions\FieldsPack\Blocks;
! defined( 'ABSPATH' ) && exit();


use TotalSurvey\Blocks\BlockType;
use TotalSurvey\Blocks\Concerns\GenerateInputName;
use TotalSurvey\Extensions\FieldsPack\Validations\MatrixValidation;
use TotalSurvey\Models\Block;
use TotalSurvey\Models\Entry;
use TotalSurvey\Models\EntryBlock;
use TotalSurveyVendors\TotalSuite\Foundation\Helpers\Html;
use TotalSurveyVendors\TotalSuite\Foundation\Support\Arrays;

class Matrix extends BlockType
{
    public static $category  = 'question';
    public static $id        = 'matrix';
    public static $icon      = 'subject';
    public static $static    = false;
    public static $aggregate = true;
    public static $aliases   = ['matrix'];

    use GenerateInputName;

    public static function render(Block $block)
    {
        $columns = $block->option('columns', []);
        $rows    = $block->option('rows', []);

        $div       = Html::create('div', ['class' => 'matrix-container']);
        $table     = Html::create(
            'table',
            ['class' => "totalsurvey-matrix {$block->option('cssClass')}", 'id' => "matrix-{$block->uid}"]
        );
        $tableHead = Html::create('tr');
        $tableBody = Html::create('tr');

        $tableHead->addContent(Html::create('th', []));
        foreach ($columns as $column):
            $tableHead->addContent(Html::create('th', [], $column['label']));
        endforeach;

        foreach ($rows as $row) :
            $tableBodyRow = Html::create('tr');
            $tableBodyRow->addContent(Html::create('td', [], $row['label']));

            foreach ($columns as $column):
                $tableBodyRow->addContent(
                    Html::create(
                        'td',
                        [],
                        Html::create(
                            'label',
                            ['class' => 'radio'],
                            Html::create(
                                'input',
                                [
                                    'type'                => 'radio',
                                    'name'                => static::getInputName(
                                        $block->section->uid,
                                        $block->uid,
                                        $row['uid']
                                    ),
                                    'value'               => $column['uid'],
                                    'data-variable-name'  => $row['uid'],
                                    'data-variable-value' => $row['label'].' - '.$column['label'],
                                ]
                            )
                        )
                    )
                );
            endforeach;

            $tableBody->addContent($tableBodyRow);

        endforeach;

        $div->addContent(
            Html::create(
                'input',
                [
                    'type'                => 'hidden',
                    'name'                => '',
                    'value'               => '',
                    'data-variable-name'  => $block->uid,
                    'data-variable-value' => '',
                ]
            )
        );


        $table->addContent(Html::create('thead', [], $tableHead));
        $table->addContent(Html::create('tbody', [], $tableBody));
        $div->addContent($table);

        return $div;
    }

    public static function getSerializedFromRawValue(Block $block, Entry $entry, $value)
    {
        return $value;
    }

    public static function getTextFromRawValue(Block $block, Entry $entry, $value)
    {
        $text    = [];
        $rows    = Arrays::extract($block->option('rows', []), 'label', 'uid');
        $columns = Arrays::extract($block->option('columns', []), 'label', 'uid');

        foreach ($rows as $rowUID => $rowLabel) {
            foreach ($columns as $columnUID => $columnLabel) {
                if (isset($value[$rowUID]) && $value[$rowUID] === $columnUID) {
                    $text[] = sprintf('%s: %s', $rowLabel, $columnLabel);

                    break;
                }
            }
        }

        return implode("\n", $text);
    }

    public static function getMetadataFromRawValue(Block $block, Entry $entry, $value)
    {
        $rows    = Arrays::extract($block->option('rows', []), 'label', 'uid');
        $columns = Arrays::extract($block->option('columns', []), 'label', 'uid');

        $table   = [];
        $choices = [];
        $table[] = array_merge([''], array_values($columns));

        foreach ($rows as $rowUID => $rowLabel) {
            $tableRow = [$rowLabel];

            foreach ($columns as $columnUID => $columnLabel) {
                $tableRow[] = isset($value[$rowUID]) && $value[$rowUID] === $columnUID ? '✔' : '';
                if (isset($value[$rowUID]) && $value[$rowUID] === $columnUID) {
                    $choices[$rowLabel] = $columnLabel;
                }
            }

            $table[] = $tableRow;
        }

        return ['table' => $table, 'choices' => $choices];
    }

    public static function getValidationRules(Block $block)
    {
        $rules = parent::getValidationRules($block);

        if ($block->field->isRequired()) {
            $rules[$block->uid]['matrix'] = new MatrixValidation($block);
        }

        return $rules;
    }

    public static function getValidationMessages(Block $block)
    {
        return [
            "{$block->uid}:matrix" => __('Please fill all criteria.', 'totalsurvey'),
        ];
    }

    public static function migrate(EntryBlock $entryBlock)
    {
        $value   = $entryBlock->value;
        $text    = [];
        $columns = $value[0];
        array_shift($columns);
        array_shift($value);

        foreach ($value as $index => $row) {
            $label = $row[0];
            array_shift($row);
            foreach ($row as $columnIndex => $column) {
                if (!empty($column)) {
                    $text[] = sprintf('%s: %s', $label, $columns[$columnIndex]);
                }
            }
        }

        $entryBlock->text  = implode("\n", $text);
        $entryBlock->value = $entryBlock->section->entryData->raw[$entryBlock->section->uid][$entryBlock->uid] ?? [];
    }

    public static function optionsFromEntryBlock(EntryBlock $entryBlock)
    {
        $rows    = array_keys($entryBlock->meta['choices']);
        $columns = array_slice($entryBlock->meta['table'][0], 1);

        return [
            'rows'    => array_map(
                static function ($row) {
                    return ['label' => $row];
                },
                $rows
            ),
            'columns' => array_map(
                static function ($column) {
                    return ['label' => $column];
                },
                $columns
            ),
        ];
    }

    public static function getTranslatableAttributes()
    {
        return [
            'field.options.rows.label',
            'field.options.columns.label',
        ];
    }
}
