<?php

namespace TotalSurvey\Extensions\Insights\Tasks;
! defined( 'ABSPATH' ) && exit();


use Exception;
use TotalSurvey\Extensions\FieldsPack\Blocks\Rating;
use TotalSurvey\Extensions\Insights\Filters\InsightsFieldsFilter;
use TotalSurvey\Extensions\Insights\Models\InsightsData;
use TotalSurvey\Models\Block;
use TotalSurvey\Models\Entry;
use TotalSurvey\Models\EntryBlock;
use TotalSurvey\Models\Survey;
use TotalSurveyVendors\TotalSuite\Foundation\Database\Query;
use TotalSurveyVendors\TotalSuite\Foundation\Database\Query\Expression;
use TotalSurveyVendors\TotalSuite\Foundation\Database\Query\Select;
use TotalSurveyVendors\TotalSuite\Foundation\Support\Arrays;
use TotalSurveyVendors\TotalSuite\Foundation\Support\Collection;
use TotalSurveyVendors\TotalSuite\Foundation\Task;

class GenerateInsightsForSurvey extends Task
{

    /**
     * @var Survey
     */
    protected $survey;

    /**
     * @var Collection|Entry[]
     */
    protected $entries;

    /**
     * @var string
     */
    protected $to;

    /**
     * @var mixed|null
     */
    protected $from;

    /**
     * @var bool
     */
    protected $includeObsolete;

    /**
     * EntriesForInsights constructor.
     *
     * @param  Survey  $survey
     * @param  null  $from
     * @param  null  $to
     */
    public function __construct(Survey $survey, $from = null, $to = null, $includeObsolete = false)
    {
        $this->survey          = $survey;
        $this->from            = $from;
        $this->to              = $to;
        $this->includeObsolete = $includeObsolete;
        $this->entries         = $this->getEntries()->get();
    }

    /**
     * @return bool|mixed|void
     */
    protected function validate()
    {
        return true;
    }

    /**
     * @return InsightsData[]
     */
    protected function execute()
    {
        $sections = $this->processSections();
        $this->normalizeValues($sections);

        return [
            'users'    => $this->getCountOfUsers(),
            'daily'    => $this->getWeeklyCount($this->survey),
            'count'    => [
                'abs'       => $this->countAllEntries(),
                'total'     => $this->entries->count(),
                'thisWeek'  => $this->getTotalThisWeek($this->survey),
                'thisMonth' => $this->getTotalThisMonth($this->survey),
                'today'     => $this->getTotalToday($this->survey),
            ],
            'sections' => array_values($sections),
        ];
    }

    /**
     * @return int
     */
    protected function countAllEntries()
    {
        return Entry::query()
                    ->where('survey_uid', $this->survey->uid)
                    ->count();
    }

    /**
     *
     * @return Query|Select
     */
    protected function getEntries()
    {
        $query = Entry::query()
                      ->where('survey_uid', $this->survey->uid)
                      ->where('status', Entry::STATUS_OPEN);


        if ($this->from !== null && $this->to !== null && $this->from === $this->to) {
            $date = date('Y-m-d H:i:s', strtotime($this->from));
            $query->where(new Expression("DATE(created_at)"), new Expression("DATE('$date')"));
        } else {
            if ($this->from !== null) {
                $query->where('created_at', '>=', date('Y-m-d H:i:s', strtotime($this->from)));
            }

            if ($this->to !== null) {
                $query->where(new Expression('created_at'), '<=', date('Y-m-d H:i:s', strtotime($this->to)));
            }
        }

        return $query;
    }

    /**
     * @return array
     */
    protected function processSections(): array
    {
        $data             = [];
        $aggregatedFields = InsightsFieldsFilter::apply(
            ['scale', 'select', 'radio', 'checkbox', 'rating', 'image-choice', 'matrix']
        );
        foreach ($this->entries as $entry) {
            foreach ($entry->data->sections as $entrySection) {
                if ($entrySection->blocks->isEmpty()) {
                    continue;
                }

                try {
                    $section = $this->survey->getSection($entrySection->uid);

                    if (!isset($data[$section->uid])) {
                        $data[$section->uid] = [
                            'uid'    => $section->uid,
                            'title'  => $section->title,
                            'blocks' => [],
                        ];
                    }

                    $index = 0;

                    foreach ($section->blocks as $block) {
                        $index++;


                        if (!$block->blockType::$aggregate) {
                            continue;
                        }

                        if (!isset($data[$section->uid]['blocks'][$block->uid])) {
                            $data[$section->uid]['blocks'][$block->uid] = [
                                'title'   => $block->title ?: "Question #{$index}",
                                'type'    => $block->field->type,
                                'values'  => [],
                                'options' => $block->field->options,
                            ];
                        }

                        $values = $data[$section->uid]['blocks'][$block->uid]['values'];

                        $entryBlock = $entrySection->blocks->where(['uid' => $block->uid])->first();

                        $userValue = $this->getValue($block, $entryBlock->value ?? null);

                        if (is_array($userValue) && $block->typeId !== 'question:matrix') {
                            $values = array_merge($userValue, $values);
                        } else {
                            $values[] = $userValue;
                        }

                        $data[$section->uid]['blocks'][$block->uid]['values'] = $values;
                    }
                    $calculatedBlocksValues = array_fill(0, count($section->blocks), true);
                    $calculatedBlocks       = array_combine(
                        Arrays::extract($section->blocks, 'uid'),
                        array_values($calculatedBlocksValues)
                    );
                } catch (Exception $e) {
                }

                if (!$this->includeObsolete) {
                    continue;
                }

                if (!isset($data[$entrySection->uid])) {
                    $data[$entrySection->uid] = [
                        'title'  => $entrySection->title.' (Obsolete)',
                        'blocks' => [],
                    ];
                }

                foreach ($entrySection->blocks as $entrySectionBlock) {
                    if (!$entrySectionBlock->blockType::$aggregate) {
                        continue;
                    }

                    if (isset($calculatedBlocks[$entrySectionBlock->uid]) && $calculatedBlocks[$entrySectionBlock->uid] == true) {
                        continue;
                    }


                    if (!(isset($data[$entrySection->uid]['blocks'][$entrySectionBlock->uid]))) {
                        $data[$entrySection->uid]['blocks'][$entrySectionBlock->uid] = [
                            'title'   => $entrySectionBlock->title.' (obsolete)',
                            'type'    => $entrySectionBlock->blockType::$id,
                            'values'  => [],
                            'options' => $entrySectionBlock->blockType::optionsFromEntryBlock($entrySectionBlock),
                        ];
                    }

                    $values = $data[$entrySection->uid]['blocks'][$entrySectionBlock->uid]['values'];

                    $userValue = $this->getObsoleteValue($entrySectionBlock);

                    if (is_array($userValue) && $entrySectionBlock->typeId !== 'question:matrix') {
                        $values = array_merge($userValue, $values);
                    } else {
                        $values[] = $userValue;
                    }

                    $data[$entrySection->uid]['blocks'][$entrySectionBlock->uid]['values'] = $values;
                }
            }
        }

        return array_filter($data);
    }

    /**
     * @param  Block  $block
     * @param $value
     *
     * @return array|mixed
     */
    protected function getValue(Block $block, $value)
    {
        if ($value === null) {
            return [];
        }

        if (in_array($block->typeId, ['question:choices', 'question:dropdown', 'question:image-choice'])) {
            $choice = Arrays::first(
                $block->field->options['choices'],
                function ($choice) use ($value) {
                    return $choice['uid'] === $value;
                }
            );

            if (empty($choice)) {
                return __('Other', 'totalsurvey');
            }

            return $choice['label'];
        }

        if ($block->typeId === 'question:multiple-choices') {
            $choices = Arrays::where(
                $block->field->options['choices'],
                function ($choice) use ($value) {
                    return in_array($choice['uid'], (array) $value, true);
                }
            );

            return Arrays::extract($choices, 'label');
        }

        if ($block->typeId === 'question:scale') {
            return $value ? sprintf("%s of %s", $value, Arrays::get($block->field->options, 'scale')) : $value;
        }

        if ($block->typeId === 'question:rating') {
            return $block->blockType::getSymbolAltValue($value);
        }

        if ($block->typeId === 'question:matrix') {
            $rows    = Arrays::get($block->field->options, 'rows');
            $columns = Arrays::get($block->field->options, 'columns');


            $values = array_fill(0, count($columns), array_fill(0, count($rows), 0));
            $data   = array_combine(Arrays::extract($columns, 'label'), array_values($values));

//            $index = 0;
            foreach ($value as $key => $v) {
                if (!empty($key)) {
                    $index                          = array_search($key, array_column($rows, 'uid'));
                    $column                         = Arrays::first(
                        $columns,
                        function ($item) use ($v) {
                            return $item['uid'] === $v;
                        }
                    );
                    $data[$column['label']][$index] += 1;
//                $index++;
                }
            }

            return $data;
        }

        return $value;
    }

    /**
     * @param  EntryBlock  $block
     * @param $value
     *
     * @return array|mixed
     */
    protected function getObsoleteValue(EntryBlock $block)
    {
        if (in_array($block->typeId, ['question:choices', 'question:dropdown', 'question:image-choice'])) {
            return $block->text;
        }

        if ($block->typeId === 'question:multiple-choices') {
            return explode(',', $block->text);
        }

        if ($block->typeId === 'question:scale') {
            return $block->text;
        }

        if ($block->typeId === 'question:rating') {
            return $block->blockType::getSymbolAltValue($block->value);
        }

        if ($block->typeId === 'question:matrix') {
            $rows    = array_map([$this, 'wrapLabel'], array_keys($block->meta['choices']));
            $columns = array_map([$this, 'wrapLabel'], array_slice($block->meta['table'][0], 1));

            $values = array_fill(0, count($columns), array_fill(0, count($rows), 0));
            $data   = array_combine(Arrays::extract($columns, 'label'), array_values($values));
            foreach (array_slice($block->meta['table'], 1) as $index => $row) {
                $columnIndex = array_search('✔', $row, true);
                if (isset($columns[$columnIndex - 1])) {
                    $column = $columns[$columnIndex - 1]['label'];
                    if (!isset($data[$column][$index])) {
                        $data[$column][$index] = 0;
                    }
                    $data[$column][$index] += 1;
                }
            }

            return $data;
        }


        return $block->value;
    }

    public function wrapLabel($item)
    {
        return ['label' => $item];
    }

    /**
     * @param  array  $sections
     */
    protected function normalizeValues(array &$sections)
    {
        foreach ($sections as $id => &$section) {
            foreach ($section['blocks'] as $index => &$block) {
                if ($block['type'] !== 'matrix') {
                    $count = array_count_values($block['values']);

                    if ($block['type'] !== 'scale' && $block['type'] !== 'rating') {
                        arsort($count, SORT_DESC);
                    } elseif ($block['type'] !== 'rating') {
                        arsort($count, SORT_ASC);
                    } elseif ($block['type'] === 'rating') {
                        $ratings              = [];
                        $block['accumulated'] = 0;

                        for ($ratingIndex = 1; $ratingIndex <= 5; $ratingIndex++) {
                            $label                = Rating::getSymbolAltValue($ratingIndex);
                            $ratings[$label]      = empty($count[$label]) ? 0 : $count[$label];
                            $block['accumulated'] += $ratings[$label] * $ratingIndex;
                        }
                        $count = $ratings;
                    }

                    $block['total'] = array_sum($count);

                    if (!empty($block['accumulated'])) {
                        $block['average'] = (double) number_format($block['accumulated'] / $block['total'], 1);
                    }
                } else {
                    list($count, $block['total']) = $this->normalizeMatrix($block);
                }

                $block['values'] = $count;

                unset($block['options']);
            }

            $section['blocks'] = array_values($section['blocks']);
        }
    }

    protected function normalizeMatrix($question)
    {
        $rows = Arrays::extract(Arrays::get($question['options'], 'rows'), 'label');

        $criteria = [];
        $data     = [];

        foreach ($question['values'] as $criterion => $values) {
            foreach ($values as $key => $value) {
                if (!isset($data[$key])) {
                    $data[$key] = $value;
                } else {
                    foreach ($value as $x => $y) {
                        $data[$key][$x] += $y;
                    }
                }
            }
        }

        $total = 0;

        foreach ($rows as $index => $row) {
            $criterion = [];

            foreach ($data as $key => $value) {
                if (!isset($criterion[$key])) {
                    $criterion[$key] = $value[$index];
                } else {
                    $criterion[$key] += $value[$index];
                }
            }

            $total += array_sum($criterion);

            $criteria[$row] = [
                'values' => $criterion,
                'total'  => array_sum($criterion),
            ];
        }

        return [['criteria' => $criteria, 'data' => $data], $total];
    }


    /**
     *
     * @return array
     */
    protected function getCountOfUsers()
    {
        $total = $this->entries->count();

        if ($total === 0) {
            return [
                'authenticated' => 0,
                'anonymous'     => 0,
            ];
        }

        $users = $this->getEntries()
                      ->where(
                          function (Select $query) {
                              $query->where('user_id', '>', 0)
                                    ->orWhere('user_id', '!=', 'NULL');
                          }
                      )
                      ->count();

        $anonymous = abs($total - $users);

        return [
            'authenticated' => round($users / $total * 100),
            'anonymous'     => round($anonymous / $total * 100),
        ];
    }

    /**
     * @param  Survey  $survey
     *
     * @return array
     */
    protected function getWeeklyCount(Survey $survey)
    {
        $expression = new Expression('DAYOFWEEK(created_at)');

        return Entry::query()
                    ->column($expression, 'day')
                    ->column(new Expression('COUNT(*)'), 'total')
                    ->where('survey_uid', $survey->uid)
                    ->where('status', Entry::STATUS_OPEN)
                    ->where(new Expression('YEARWEEK(NOW())'), new Expression('YEARWEEK(created_at)'))
                    ->groupBy($expression)
                    ->execute();
    }

    /**
     * @param  Survey  $survey
     *
     * @return int
     */
    protected function getTotalThisWeek(Survey $survey)
    {
        return (int) Entry::query()
                          ->where('survey_uid', $survey->uid)
                          ->where('status', Entry::STATUS_OPEN)
                          ->where(new Expression('WEEK(created_at)'), new Expression('WEEK(CURRENT_DATE)'))
                          ->count();
    }

    /**
     * @param  Survey  $survey
     *
     * @return int
     */
    protected function getTotalThisMonth(Survey $survey)
    {
        return (int) Entry::query()
                          ->where('survey_uid', $survey->uid)
                          ->where('status', Entry::STATUS_OPEN)
                          ->where(new Expression('MONTH(created_at)'), new Expression('MONTH(CURRENT_DATE)'))
                          ->count();
    }

    /**
     * @param  Survey  $survey
     *
     * @return int
     */
    protected function getTotalToday(Survey $survey)
    {
        return Entry::query()
                    ->where('survey_uid', $survey->uid)
                    ->where('status', Entry::STATUS_OPEN)
                    ->where(new Expression('DATE(created_at)'), new Expression('DATE(CURRENT_DATE)'))
                    ->count();
    }
}
