<?php

namespace TotalSurvey\Extensions\Limitations;
! defined( 'ABSPATH' ) && exit();


use TotalSurvey\Events\Entries\OnPostValidateEntry;
use TotalSurvey\Events\Surveys\OnResetSurvey;
use TotalSurvey\Extensions\Limitations\Exception\FinishedException;
use TotalSurvey\Extensions\Limitations\Exception\NotStartedException;
use TotalSurvey\Extensions\Limitations\Strategies\CookieLimitationStrategy;
use TotalSurvey\Extensions\Limitations\Strategies\IPLimitationStrategy;
use TotalSurvey\Extensions\Limitations\Strategies\UserLimitationStrategy;
use TotalSurvey\Extensions\Limitations\Tasks\CheckLimitations;
use TotalSurvey\Filters\Entries\EntryCreatedResponseFilter;
use TotalSurvey\Filters\Surveys\SurveyPreRenderFilter;
use TotalSurvey\Models\Entry;
use TotalSurvey\Models\Survey;
use TotalSurveyVendors\TotalSuite\Foundation\Exceptions\Exception;
use TotalSurveyVendors\TotalSuite\Foundation\Helpers\Html;
use TotalSurveyVendors\TotalSuite\Foundation\Http\Request;
use TotalSurveyVendors\TotalSuite\Foundation\Http\Response;
use TotalSurveyVendors\TotalSuite\Foundation\WordPress\Modules\Extension;

class Module extends Extension
{
    /**
     * @var LimitationStrategy[]
     */
    protected $limitations = [];


    /**
     * @return mixed|void
     */
    public function run()
    {
        $request = Request::instance();

        $this->limitations[] = new CookieLimitationStrategy($request);
        $this->limitations[] = new IPLimitationStrategy($request);
        $this->limitations[] = new UserLimitationStrategy($request);

        SurveyPreRenderFilter::add([$this, 'check']);
        EntryCreatedResponseFilter::add([$this, 'apply'], 10, 3);

        OnResetSurvey::listen([$this, 'onResetSurvey']);
        OnPostValidateEntry::listen([$this, 'validate']);

        // Frontend
        SurveyPreRenderFilter::add([$this, 'render']);

        // Backend
        OnPostValidateEntry::listen(
            function (OnPostValidateEntry $event) {
                $this->checkSettings($event->survey);
            }
        );
    }

    /**
     * @param  OnPostValidateEntry  $event
     *
     * @throws Exception
     */
    public function validate(OnPostValidateEntry $event)
    {
        if ($event->survey->isLimitationEnabled('entry')) {
            foreach ($this->limitations as $limitation) {
                $limitation->check($event->survey);
            }
        }
    }

    /**
     * @param $can
     * @param  Survey  $survey
     *
     * @return mixed
     * @throws Exception
     */
    public function canSubmitEntry($can, Survey $survey)
    {
        return CheckLimitations::invoke($this->limitations, $survey);
    }

    /**
     * @param  OnResetSurvey  $event
     */
    public function onResetSurvey(OnResetSurvey $event)
    {
        foreach ($this->limitations as $limitation) {
            $limitation->reset($event->survey);
        }
    }

    /**
     * @param  Survey|Html  $survey
     *
     * @return string|Html|null
     * @throws Exception
     */
    public function check($survey)
    {
        if ($survey instanceof Survey && $survey->isLimitationEnabled('entry')) {
            $survey->setAttribute('canRestart', CheckLimitations::invoke($this->limitations, $survey));
        }

        return $survey;
    }

    /**
     * @param  Response  $response
     * @param  Survey  $survey
     * @param  Entry  $entry
     *
     * @return Response
     * @throws Exception
     */
    public function apply(Response $response, Survey $survey, Entry $entry)
    {
        foreach ($this->limitations as $limitation) {
            $response = $limitation->apply($response, $survey, $entry);
        }

        $entry = $entry->toPublic();
        $entry->setAttribute('canRestart', CheckLimitations::invoke($this->limitations, $survey, true));

        return $response->withJson(
            [
                'success' => true,
                'data'    => $entry->toArray(),
            ]
        );
    }

    /**
     * @param  Survey  $survey
     *
     * @return Html|Survey
     */
    public function render(Survey $survey)
    {
        $this->checkSettings($survey);

        return $survey;
    }

    /**
     * @param  Survey  $survey
     *
     * @return bool
     * @throws Exception
     */
    public function checkSettings(Survey $survey)
    {
        if ($survey->getSettings('limitations.date.enabled', false)) {
            $now   = mysql2date('U', 'now');
            $start = $survey->getSettings('limitations.date.start', null);
            $end   = $survey->getSettings('limitations.date.end', null);

            NotStartedException::throwIf(
                $start && strtotime($start, $now) > $now,
                $survey->getSettings('limitations.date.notStartedMessage') ?: __(
                    'This survey has not started yet.',
                    'totalsuite'
                )
            );

            FinishedException::throwIf(
                $end && strtotime($end, $now) <= $now,
                $survey->getSettings('limitations.date.finishedMessage') ?: __('This survey has ended.', 'totalsuite')
            );
        }

        return true;
    }
}
