<?php

namespace TotalSurvey\Extensions\Scoring;
! defined( 'ABSPATH' ) && exit();


use TotalSurvey\Extensions\Scoring\Model\Scoring;
use TotalSurvey\Extensions\Scoring\Tasks\CalculateEntryScore;
use TotalSurvey\Filters\Blocks\BlockMentionFilter;
use TotalSurvey\Filters\Entries\BeforeEntrySaveFilter;
use TotalSurvey\Filters\Entries\EntriesExportHeadersFilter;
use TotalSurvey\Filters\Entries\EntriesExportRowFilter;
use TotalSurvey\Filters\Entries\EntriesQueryFilter;
use TotalSurvey\Filters\Entries\EntryToExportDataFilter;
use TotalSurvey\Filters\Entries\PublicEntryFilter;
use TotalSurvey\Filters\Surveys\SurveyAssetsFilter;
use TotalSurvey\Filters\Surveys\SurveySettingsFilter;
use TotalSurvey\Models\Entry;
use TotalSurvey\Models\Survey;
use TotalSurvey\Plugin;
use TotalSurveyVendors\TotalSuite\Foundation\Database\Query;
use TotalSurveyVendors\TotalSuite\Foundation\Helpers\Html;
use TotalSurveyVendors\TotalSuite\Foundation\Support\Arrays;
use TotalSurveyVendors\TotalSuite\Foundation\Support\Strings;
use TotalSurveyVendors\TotalSuite\Foundation\WordPress\Modules\Extension;

class Module extends Extension
{
    public function run()
    {
        // Fields Custom CSS
        SurveyAssetsFilter::add([$this, 'assets']);

        // Filter survey settings
        SurveySettingsFilter::add([$this, 'filterSettings']);

        // Apply Scoring
        BeforeEntrySaveFilter::add([$this, 'applyScoring'], 10, 2);

        // Filter entries
        EntriesQueryFilter::add([$this, 'filterEntries'], 10, 3);

        // Filter CSV export headers
        EntriesExportHeadersFilter::add([$this, 'filterExportHeaders'], 10, 2);

        // Filter CSV export row
        EntriesExportRowFilter::add([$this, 'filterExportRow'], 10, 4);

        // Filter CSV data
        EntryToExportDataFilter::add([$this, 'filterExportData'], 10, 3);
    }

    /**
     * @param  array  $data
     * @param  array  $entry
     * @param  bool  $format
     *
     * @return array
     */
    public function filterExportData(array $data, array $entry, $format)
    {
        if ($format === 'csv') {
            $data['scoring'] = Arrays::get($entry, 'data.scoring', []);
        }

        return $data;
    }


    /**
     * @param  array  $settings
     *
     * @return array
     */

    public function filterSettings(array $settings)
    {
        if (!empty($settings['scoring']['grades'])) {
            foreach ($settings['scoring']['grades'] as $index => $grade) {
                if (is_string($grade['content'])) {
                    $compiledBlocks = [];
                    if (!empty($grade['content'])) {
                        $raw   = $grade['content'];
                        $texts = preg_split("/\{\{.+?\}\}/", $raw);
                        preg_match_all("/\{\{.+?\}\}/", $raw, $matches);
                        foreach ($texts as $textIndex => $text) {
                            $compiledBlocks[] = ['text' => $text];
                            if (isset($matches[0][$textIndex])) {
                                $compiledBlocks[] = [
                                    'type' => 'mention',
                                    '_uid' => 'scoring:'.trim(
                                            str_replace(['{{', '}}'], '', $matches[0][$textIndex])
                                        ),

                                    'label'    => str_replace(['{{', '}}'], '', $matches[0][$textIndex]),
                                    'fallback' => "",
                                    'children' => [['text' => '']],
                                ];
                            }
                        }
                    }

                    $newContent                                       = [
                        "blocks" => [
                            [
                                'type'         => 'content',
                                'typeId'       => 'content:paragraph',
                                'uid'          => Strings::uid(),
                                'category'     => 'content',
                                'translations' => [],
                                'content'      => [
                                    'value'     => $grade['content'],
                                    'options'   => [
                                        'alignment' => 'start',
                                    ],
                                    'type'      => 'paragraph',
                                    'value_raw' => [['type' => 'paragraph', "children" => $compiledBlocks]],
                                ],
                            ],
                        ],

                    ];
                    $settings['scoring']['grades'][$index]['content'] = $newContent;
                    $settings['scoring']['grades'][$index]['uid']     = $settings['scoring']['grades'][$index]['uid'] ?? Strings::uid(
                    );
                }
            }
        }

        return $settings;
    }

    /**
     * @param  array  $values
     * @param  array  $entry
     * @param  string  $format
     *
     * @return array
     */
    public function filterExportRow(array $values, array $entry, $format)
    {
        if ($format === 'csv') {
            $values[] = Arrays::get($entry, 'scoring.score', '0');
            $values[] = Arrays::get($entry, 'scoring.percentage', '0').'%';
            $values[] = Arrays::get($entry, 'scoring.grade.label', 'N/A');
        }

        return $values;
    }

    /**
     * @param  array  $headers
     * @param  string  $format
     *
     * @return array
     */
    public function filterExportHeaders(array $headers, $format)
    {
        if ($format === 'csv') {
            $headers['score']      = __('Score', 'totalsurvey');
            $headers['percentage'] = __('Percentage', 'totalsurvey');
            $headers['grade']      = __('Grade', 'totalsurvey');
        }

        return $headers;
    }

    /**
     * @param  Query  $query
     * @param  array  $filters
     *
     * @return Query
     */
    public function filterEntries(Query $query, array $filters)
    {
        if (!empty($filters['score_min'])) {
            $query->where(Query::raw('FLOOR((score / total) * 100 )'), '>=', (int) $filters['score_min']);
        }

        if (!empty($filters['score_max'])) {
            $query->where(Query::raw('FLOOR((score / total) * 100 )'), '<=', (int) $filters['score_max']);
        }

        return $query;
    }

    /**
     * @param  Entry  $entry
     * @param  Survey  $survey
     *
     * @return Entry
     * @throws \TotalSurveyVendors\TotalSuite\Foundation\Exceptions\Exception
     */
    public function applyScoring($entry, $survey)
    {
        if (!$survey->getSettings('scoring.enabled', false)) {
            return $entry;
        }

        $scoring = Scoring::from($survey->getSettings('scoring', []));

        if ($scoring->score === 0) {
            return $entry;
        }

        $entryScore = CalculateEntryScore::invoke($entry, $scoring);

        $entry->setAttribute('score', $entryScore->score);
        $entry->setAttribute('total', (int) $entryScore->total);
        $entry->setAttribute('data.scoring', $entryScore);

        BlockMentionFilter::add(
            function (string $label, $part) use ($entryScore) {
                return $this->getMentionLabel($label, $part, $entryScore);
            },
            10,
            2
        );

        PublicEntryFilter::add(static function (Entry $entry) use ($survey, $entryScore) {
            $entry->setAttribute('score', $entryScore);
            $entry->setAttribute(
                'customThankYouMessage',
                $entryScore->grade->getCustomThankYouMessage($entryScore, $survey)
            );

            return $entry;
        });


        return $entry;
    }

    /**
     * @param $assets
     *
     * @return array
     */
    public function assets($assets)
    {
        $assets['css']['scoring'] = $this->getUrl('assets/style.css?ver=' . Plugin::env('version'));

        return $assets;
    }

    private function getMentionLabel(string $label, $part, Model\EntryScore $entryScore)
    {
        if (str_starts_with($part['_uid'], 'scoring:score')) {
            $label = $entryScore->score;
        } elseif (str_starts_with($part['_uid'], 'scoring:grade')) {
            $label = $entryScore->grade->label;
        } elseif (str_starts_with($part['_uid'], 'scoring:percentage')) {
            $label = $entryScore->percentage;
        } elseif (str_starts_with($part['_uid'], 'scoring:total')) {
            $label = $entryScore->total;
        } elseif (str_starts_with($part['_uid'], 'scoring:badge')) {
            $label = (string) Html::create(
                'span',
                [
                    'class' => 'scoring-badge',
                    'style' => "background-color: {$entryScore->grade->color}",
                ],
                $entryScore->grade->label
            );
        } else {
            // when variable
            $label = "{{ variable('{$part['_uid']}', '{$part['fallback']}') }}";
        }

        // Only REST context (Display text instead of variable notation)

        return $label;
    }
}
