<?php

namespace TotalSurvey\Extensions\Scoring\Tasks;
! defined( 'ABSPATH' ) && exit();


use TotalSurvey\Extensions\Scoring\Model\EntryScore;
use TotalSurvey\Extensions\Scoring\Model\Scoring;
use TotalSurvey\Extensions\Scoring\Model\ScoringBlock;
use TotalSurvey\Models\Entry;
use TotalSurvey\Tasks\ProcessCondition;
use TotalSurveyVendors\TotalSuite\Foundation\Exceptions\Exception;
use TotalSurveyVendors\TotalSuite\Foundation\Task;

/**
 * Class CalculateEntryScore
 *
 * @package TotalSurvey\Extensions\Scoring\Tasks
 * @method static EntryScore invoke(Entry $entry, Scoring $scoring)
 * @method static EntryScore invokeWithFallback($fallback, Entry $entry, Scoring $scoring)
 */
class CalculateEntryScore extends Task
{
    /**
     * @var Entry
     */
    protected $entry;

    /**
     * @var Scoring
     */
    protected $scoring;

    /**
     * Reset constructor.
     *
     * @param  Entry  $entry
     * @param  Scoring  $scoring
     */
    public function __construct(Entry $entry, Scoring $scoring)
    {
        $this->entry   = $entry;
        $this->scoring = $scoring;
    }


    /**
     * @inheritDoc
     */
    protected function validate()
    {
        return true;
    }

    /**
     * @inheritDoc
     * @throws Exception
     */
    protected function execute()
    {
        $score    = 0;
        $sections = [];

        foreach ($this->entry->data->sections as $entrySection) {
            $scoringSection = $this->scoring->sections->where(['uid' => $entrySection->uid])->first();

            if ($scoringSection === null) {
                continue;
            }

            $sectionTotal  = 0;
            $sectionBlocks = [];

            foreach ($entrySection->blocks as $entryBlock) {
                /**
                 * @var ScoringBlock $scoringBlock
                 */
                $scoringBlocks = $scoringSection->blocks->where(['uid' => $entryBlock->uid]);
                if (count($scoringBlocks) > 0) {
                    foreach ($scoringBlocks as $scoringBlockIndex => $scoringBlock) {
                        if ($scoringBlock === null) {
                            continue;
                        }


                        $sectionBlocks[$scoringBlock->uid] = $sectionBlocks[$scoringBlock->uid] ?? 0;

                        foreach ($scoringBlock->rules as $ruleIndex => $rule) {
                            $matching = ProcessCondition::invoke(
                                $entryBlock->value,
                                $rule->operator,
                                $rule->value
                            );

                            $ruleScore = $matching ? $rule->score : 0;

                            if (count($scoringBlocks) > 1) {
                                $sectionBlocks[$scoringBlock->uid] += $ruleScore;
                            } elseif ($ruleScore > $sectionBlocks[$scoringBlock->uid]) {
                                $sectionBlocks[$scoringBlock->uid] = $ruleScore;
                            }
                        }
                    }

                    $sectionTotal += $sectionBlocks[$scoringBlock->uid];
                }
            }

            $score                        += $sectionTotal;
            $sections[$entrySection->uid] = [
                'score'  => $sectionTotal,
                'blocks' => $sectionBlocks,
            ];
        }

        $percentage = ceil(($score / $this->scoring->score) * 100);
        $grade      = $this->scoring->getEntryGrade($percentage);

        return EntryScore::from(
            [
                'total'      => $this->scoring->score,
                'score'      => $score,
                'sections'   => $sections,
                'percentage' => $percentage,
                'grade'      => $grade,
            ]
        );
    }
}
